import { FormDataFromObject } from "~/lib/form_data";

declare global {
  interface Window {
    dataLayer?: unknown[];
    gtag?: unknown;
  }
}

import type { FormSystemName } from "~/types/form_system_name";

import type { ParentComponent } from "solid-js";
import { createContext, useContext } from "solid-js";
import { createStore } from "solid-js/store";
import { useWindowSize } from "@solid-primitives/resize-observer";
import { localStorageKnowsIfAlreadySubmittedForNid } from "~/lib/form_already_submitted_localstorage";
import { isFormSystemName } from "~/types/form_system_name";
import { createLocalStorage } from "@solid-primitives/storage";

type ClickEventType = MouseEvent & {
  currentTarget: HTMLButtonElement;
  target: Element;
};

type EventOptions = {
  nid?: number;
  sendToBackend?: boolean;
};

export const isClickEventType = (
  event: ClickEventType | EventZone,
): event is ClickEventType => {
  return typeof event === "object" && "target" in event;
};

type EventsContextStore = [
  {
    form?: FormSystemName;
    template?: EventTemplate;
    category?: EventCategory;
    label?: string;
    step?: number;
    gaEnabled: boolean;
    gtmEnabled: boolean;
    piwikProEnabled: boolean;
  },
  {
    setForm?: (form: FormSystemName) => void;
    setTemplate?: (template: EventTemplate) => void;
    setCategory?: (category: EventCategory) => void;
    setLabel?: (label: string) => void;
    setStep?: (number: number) => void;
    setGaEnabled?: (enabled: boolean) => void;
    setGtmEnabled?: (enabled: boolean) => void;
    setPiwikProEnabled?: (enabled: boolean) => void;
    addShowForInPageForm?: (form: FormSystemName) => void;
    sendClick?: (
      click: EventClickName,
      zone: EventZone,
      options?: EventOptions,
    ) => void;
    /**
     * Send a show event to Analytics Provider.
     *
     * @param {FormSystemName} form
     * @param {ClickEventType} | EventZone eventOrZone
     * @param {EventOptions} options
     *   - sendToBackend: Send event to backend
     *   - nid: Node id to check if form has already been submitted
     */
    sendShowEvent?: (
      form: FormSystemName,
      eventOrZone: ClickEventType | EventZone,
      options?: EventOptions,
    ) => void;
    sendSubmitEvent?: (options?: Pick<EventOptions, "sendToBackend">) => void;
    sendStepEvent?: (
      step: number,
      options?: Pick<EventOptions, "sendToBackend">,
    ) => void;
  },
];

const EventsContext = createContext<EventsContextStore>([
  {
    form: undefined,
    template: undefined,
    category: undefined,
    step: undefined,
    gaEnabled: undefined,
    gtmEnabled: undefined,
    piwikProEnabled: undefined,
  },
  {},
]);

export const EventsContextProvider: ParentComponent<{
  form?: FormSystemName;
  template?: EventTemplate;
  category?: EventCategory;
  label?: string;
  step?: number;
  gaEnabled: boolean;
  gtmEnabled: boolean;
  piwikProEnabled?: boolean;
  inPageShowAlreadyTriggered?: Set<FormSystemName>;
}> = (props) => {
  const [localStore, localStoreSet] = createLocalStorage();
  const [state, setState] = createStore({
      form: props.form || null,
      template: props.template || null,
      step: props.step || null,
      category: props.category || null,
      label: props.label || null,
      gaEnabled: props.gaEnabled || false,
      gtmEnabled: props.gtmEnabled || false,
      piwikProEnabled: localStore.piwikProEnabled
        ? parseInt(localStore.piwikProEnabled) === 1
        : true, // By default, we enable Piwik Pro tracking
    }),
    store: EventsContextStore = [
      state,
      {
        setForm(form: FormSystemName) {
          setState("form", form);
        },
        setTemplate(template: EventTemplate) {
          setState("template", template);
        },
        setStep(number: number) {
          setState("step", number);
        },
        setCategory(category: EventCategory) {
          setState("category", category);
        },
        setLabel(label: string) {
          setState("label", label);
        },
        setGaEnabled(enabled: boolean) {
          setState("gaEnabled", enabled);
        },
        setGtmEnabled(enabled: boolean) {
          setState("gtmEnabled", enabled);
        },
        setPiwikProEnabled(enabled: boolean) {
          setState("piwikProEnabled", enabled);
          localStoreSet("piwikProEnabled", enabled ? "1" : "0");
          window.dataLayer.push({
            event: "piwikPro_consent",
            piwikProAllow: enabled,
          });
        },
        sendClick(
          click: EventClickName,
          zone: EventZone,
          options: EventOptions = { sendToBackend: true },
        ) {
          const label = getEventLabel(click, state.template, zone);

          if (import.meta.env.VITE_KB_ENV !== "prod") {
            const s = "background-color: #023047; color: white";
            console.groupCollapsed(`🎯️ Click:  %c ${label} `, s);
          }

          // Google Analytics
          if (state.gaEnabled && typeof window.gtag === "function") {
            const obj = {
              event_category: "click",
              event_label: label,
            };

            if (import.meta.env.VITE_KB_ENV !== "prod") {
              console.debug("GTag event click", obj);
            }
            window.gtag("event", "click", obj);
          }

          // Google Tag Manager
          if (
            /*state.gtmEnabled && */ typeof window.dataLayer !== "undefined"
          ) {
            const obj = {
              event: "eventGA",
              category: "click",
              action: "click",
              label: label,
            };
            if (import.meta.env.VITE_KB_ENV !== "prod") {
              console.debug("GTM sending", obj);
            }
            window.dataLayer.push(obj);
          }

          if (import.meta.env.VITE_KB_ENV !== "prod") {
            console.groupEnd();
          }

          if (options.sendToBackend) {
            sendEventToBackend("click", "click", label);
          }
        },
        sendShowEvent(
          form: FormSystemName,
          eventOrZone: ClickEventType | EventZone,
          options: EventOptions = { sendToBackend: true },
        ) {
          let zone: EventZone | string;
          if (eventOrZone && isClickEventType(eventOrZone)) {
            zone = getGaZones(eventOrZone.target as HTMLElement);
          } else if (typeof eventOrZone === "string") {
            zone = eventOrZone;
          } else {
            zone = "unknown";
          }

          const label = getEventLabel(form, state.template, zone, options.nid);
          const category = getEventCategory(form);
          setState("form", form);
          setState("label", label);
          setState("category", category);

          sendShowDo(
            state,
            form,
            category,
            zone,
            options.nid,
            options.sendToBackend,
          );
        },
        sendSubmitEvent(
          options: Pick<EventOptions, "sendToBackend"> = {
            sendToBackend: true,
          },
        ) {
          if (import.meta.env.VITE_KB_ENV !== "prod") {
            const s = "background-color: #e76f51; color: white";
            console.groupCollapsed(`⛱ ️Submit: %c ${state.label} `, s);
          }

          // Google Analytics
          if (state.gaEnabled && typeof window.gtag === "function") {
            const obj = {
              event_category: state.category,
              event_label: state.label,
            };

            if (import.meta.env.VITE_KB_ENV !== "prod") {
              console.debug("GTag event submit", obj);
            }
            window.gtag("event", "submit", obj);
          }

          // Google Tag Manager
          if (
            /*state.gtmEnabled && */ typeof window.dataLayer !== "undefined"
          ) {
            console.log("dataLayer", window.dataLayer);
            const obj = {
              event: "eventGA",
              category: state.category,
              action: "submit",
              label: state.label,
            };
            if (import.meta.env.VITE_KB_ENV !== "prod") {
              console.debug("GTM sending", obj);
            }
            window.dataLayer.push(obj);
          }

          if (import.meta.env.VITE_KB_ENV !== "prod") {
            console.groupEnd();
          }

          if (options.sendToBackend) {
            sendEventToBackend("submit", state.category, state.label);
          }
        },
        sendStepEvent(
          step: number,
          options: Pick<EventOptions, "sendToBackend"> = {
            sendToBackend: true,
          },
        ) {
          const stepLabel = "step-" + step;
          const label = [
            state.form,
            stepLabel,
            state.template,
            getCurrentDeviceBasedOnViewport(),
          ].join("_");

          if (import.meta.env.VITE_KB_ENV !== "prod") {
            const s = "background-color: #bde0fe; color: black";
            console.groupCollapsed(`🦶 ️Step: %c ${label} `, s);
          }

          // Google Analytics
          if (state.gaEnabled && typeof window.gtag === "function") {
            const obj = {
              event_category: state.category,
              event_label: label,
            };

            if (import.meta.env.VITE_KB_ENV !== "prod") {
              console.debug("GTag event submit", obj);
            }
            window.gtag("event", "step", obj);
          }

          // Google Tag Manager
          if (/*state.gtmEnabled && */ typeof window.dataLayer === "object") {
            const obj = {
              event: "eventGA",
              category: state.category,
              action: "step",
              label: label,
            };
            if (import.meta.env.VITE_KB_ENV !== "prod") {
              console.debug("GTM sending", obj);
            }
            window.dataLayer.push(obj);
          }

          if (import.meta.env.VITE_KB_ENV !== "prod") {
            console.groupEnd();
          }

          if (options.sendToBackend) {
            sendEventToBackend("step", state.category, label);
          }
        },
      },
    ];

  return (
    <EventsContext.Provider value={store}>
      {props.children}
    </EventsContext.Provider>
  );
};

export function useEventsContext() {
  return useContext(EventsContext);
}

/**
 * Returns the ga-zone tree for an element.
 *
 * Returns `unknown` if the element is not in a ga-zone
 *
 * @param  {Element} elem Starting element
 * @return {String} Zone identifier
 */
function getGaZones(elem: HTMLElement): string {
  const zones = [];

  // Get the closest match
  for (; elem && elem !== document.documentElement; elem = elem.parentElement) {
    if (elem.matches("[data-ga-zone]")) {
      zones.push(elem.getAttribute("data-ga-zone"));
    }
  }

  if (zones.length === 0) {
    return "unknown";
  }

  return zones.reverse().join("-");
}

type GetEventLabelFn = (
  element: FormSystemName | EventClickName,
  template: string,
  zone: EventZone | string,
  nid?: number,
) => string;

const getEventLabel: GetEventLabelFn = (
  element,
  template,
  zone,
  nid = null,
) => {
  if (template === "search") {
    // @TODO re-implement me
    // if (window.app.searchResultsViewList) {
    //   template = 'search-list';
    // }
    //
    // if (window.app.searchResultsViewMap) {
    //   template = 'search-map';
    // }
    // // @ts-ignore
    // if (window.app.searchResultsViewMixed) {
    //   template = 'search-mix';
    // }
  }

  const device = getCurrentDeviceBasedOnViewport();

  let already = false;
  if (isFormSystemName(element)) {
    already = localStorageKnowsIfAlreadySubmittedForNid(element, nid);
  }

  if (already && ["blueprint", "booklet"].includes(element)) {
    element += "-hidden";
  }

  return [element.replace("_", "-"), template, zone, device].join("_");
};

/**
 * Sends the show event.
 *
 * @param state
 * @param form
 * @param category
 * @param zone
 * @param nid
 * @param sendToBackend
 */
function sendShowDo(
  state: { template?: string; gaEnabled: boolean; gtmEnabled: boolean },
  form: FormSystemName,
  category: EventCategory,
  zone: EventZone | string,
  nid?: number,
  sendToBackend?: boolean,
): void {
  const template = state.template ? state.template : "unknown";
  const label = getEventLabel(form, template, zone, nid);

  if (import.meta.env.VITE_KB_ENV !== "prod") {
    const s = "background-color: #ffb703; color: white";
    console.groupCollapsed(`🧘 Show:   %c ${label} `, s);
  }

  if (state.gaEnabled && typeof window.gtag === "function") {
    const obj = {
      event_category: category,
      event_label: label,
    };

    if (import.meta.env.VITE_KB_ENV !== "prod") {
      console.debug("GA sending", obj);
    }
    window.gtag("event", "show", obj);
  }

  if (/*state.gtmEnabled && */ typeof window.dataLayer === "object") {
    const obj = {
      event: "eventGA",
      category: category,
      action: "show",
      label: label,
    };
    if (import.meta.env.VITE_KB_ENV !== "prod") {
      console.debug("GTM sending", obj);
    }
    window.dataLayer.push(obj);
  }

  if (import.meta.env.VITE_KB_ENV !== "prod") {
    console.groupEnd();
  }

  if (sendToBackend) {
    sendEventToBackend("show", category, label);
  }
}

/**
 * Returns label based on the current size of the viewport.
 */
function getCurrentDeviceBasedOnViewport(): "mobile" | "tablet" | "desktop" {
  const size = useWindowSize();

  if (size.width < 768) {
    return "mobile";
  } else if (size.width < 1024) {
    return "tablet";
  }

  return "desktop";
}

function getEventCategory(form: FormSystemName): EventCategory {
  switch (form) {
    case "booklet":
    case "blueprint":
      return "download";
    case "appointment":
    case "appointment-mci":
    case "callback":
    case "parking":
    case "webinar":
    case "landing":
    case "landing-captation":
      return "meeting";
    case "program":
      return "preview";
    case "simulator-pinel":
    case "simulator-ptz":
    case "pension":
      return "simulator";
    case "configurator":
    case "mci":
    case "qualification-dwell":
    case "qualification-invest":
      return "service";
    case "cnat":
      return "cnat";
    case "contact":
    default:
      return "other";
  }
}

function sendEventToBackend(
  action: EventAction,
  category: EventCategory,
  label: string,
) {
  const url = `${import.meta.env.VITE_KB_BACKEND_BASE_URL}/es`;
  fetch(url, {
    method: "POST",
    body: FormDataFromObject({
      action: action,
      category: category,
      label: label,
    }),
  })
    .then((response) => response.json())
    .then(function (data) {
      if (import.meta.env.VITE_KB_ENV !== "prod") {
        console.groupCollapsed("🧟‍️️ Sending event to server");
        console.debug(data);
        console.groupEnd();
      }
    })
    .catch(function (error) {
      console.groupCollapsed("🧟‍ Sending event to server");
      console.error(error);
      console.groupEnd();
    });
}

export type EventAction = "click" | "show" | "submit" | "step";

export type EventTemplate =
  | "home"
  | "page-corporate"
  | "committee"
  | "landing-captation"
  | "landing-offer"
  | "landing"
  // Node bundles
  // Programs
  | "program-std"
  | "program-preview"
  | "program-out-of-stock"
  | "geography-0"
  | "geography-1"
  | "geography-2"
  | "geography-3"
  | "tag"
  | "article"
  | "search"
  | "geographies"
  | "error"
  | "simulator-pinel"
  | "simulator-ptz"
  | "jobs"
  | "articles"
  | "news-list"
  | "interviews"
  | "publications"
  | "agenda"
  | "parkings"
  | "appointment"
  | "pocs"
  | "page"
  | "plot"
  | "contacts"
  | "posts"
  | "post"
  | "unknown";

export type EventCategory =
  | "download"
  | "meeting"
  | "preview"
  | "simulator"
  | "service"
  | "cnat"
  | "click"
  | "other";

export type EventClickName =
  | "btn-search-open"
  | "btn-criterion-edit"
  | "agencies"
  | "phone"
  | "ic-open-map"
  | "btn-landing"
  | "btn-search-execute"
  | "ic-video"
  | "ic-gallery"
  | "ic-map"
  | "ic-3d"
  | "ic-magazine"
  | "ic-leaflet"
  | "ic-document"
  | "ic-block"
  | "ic-ground_plan"
  | "ic-comparator"
  | "btn-3d-visit"
  | "btn-offer"
  | "btn-send-email"
  | "btn-simulate-pinel"
  | "btn-simulate-ptz"
  | "already-owner"
  | "step-1"
  | "step-2"
  | "step-3"
  | "step-4"
  | "step-5"
  | "step-6"
  | "step-7"
  | "step-8"
  | "card"
  | "btn-see-all"
  | "btn-simulate-notary"
  | "btn-simulate-loan"
  | "btn-simulate-saving-cap"
  | "btn-simulate-pension"
  | "ic-callback-phone"
  | "logo"
  | "btn-blueprint"
  | "btn-load-map"
  | "btn-programs"
  | "anchor-grid"
  | "anchor-map"
  | "anchor-3d-visit"
  | "switch-dwell"
  | "switch-invest"
  | "filters"
  | "ic-drawer"
  | "catalog-houses"
  | "ic-blueprint-2d"
  | "ic-blueprint-3d"
  | "bc-0"
  | "bc-1"
  | "bc-2"
  | "bc-3"
  | "bc-4"
  | "bc-5"
  | "btn-leled"
  | "btn-3d-blueprint"
  | "btn-tailored"
  | "contacts"
  | `promo-${number}`;

export type EventZone =
  | "3D"
  | "breadcrumb"
  | "configurator"
  | "contact"
  | "content"
  | "drawer-dwell"
  | "drawer-invest"
  | "first-screen"
  | "footer"
  | "form"
  | "form-in-page"
  | "grid-dwell"
  | "grid-invest"
  | "grid-summary"
  | "header"
  | "header-sticky"
  | "image"
  | "in-page"
  | "intro"
  | "invest"
  | "leled"
  | "lightbox"
  | "list-card"
  | "map"
  | "map-card"
  | "media"
  | "menu-dwell"
  | "menu-invest"
  | "menu-advices"
  | "menu-group"
  | "offer"
  | "other-programs"
  | "poc"
  | "question"
  | "related-programs"
  | "result"
  | "sales-office"
  | "tools"
  | "topbar"
  | "unknown"
  | "webinar"
  | "programs-preview"
  | "programs-delivery"
  | "programs-opportunity"
  | "programs-land"
  | "programs-house";
