/* @flow strict
 *
 * TagCache
 * Allows one to cache-clear-tag data with tags and invalidate based on them
 *
 * Under the hood this stores one set of keys per tag (ID) and the data per key in Redis:
 * - `tags:foo-123` = `db.table('threads').get('foo-345') db.table('threads').get('foo-234') ...`
 * - `data:db.table('threads').get('foo-345')` = `{ "id": "foo-345", "content": { "title": "Hello" }, ... }`
 */

import Redis, { type RedisOptions } from "ioredis";

export type CacheData = unknown;

export type Options = {
  defaultTimeout?: number;
  redis?: RedisOptions;
};

class TagCache {
  redis: Redis;
  options: Options;

  constructor(port: number, host: string, options: Options = {}) {
    this.redis = new Redis(port, host, options.redis || {});
    this.options = options;
  }

  get = async (...keys: Array<string>): Promise<CacheData> => {
    try {
      const data = this.redis
        .mget(keys.map((key) => `data:${key}`))
        .then((res) => {
          try {
            // Special case for single element gets
            if (res.length === 1) return JSON.parse(res[0]);
            return res.map((elem) => JSON.parse(elem));
          } catch (err) {
            return res;
          }
        });

      this.redis.quit();
      return data;
    } catch (err) {
      return Promise.reject(err);
    }
  };

  set = async (
    key: string,
    data: CacheData,
    tags: Array<string>,
    options: {
      timeout?: number;
    } = {},
  ): Promise<unknown> => {
    try {
      // NOTE: This is a multi execution because if any of the commands is invalid
      // we don't want to execute anything
      const multi = await this.redis.multi();

      // Add the key to each of the tag sets
      tags.forEach((tag) => {
        multi.sadd(`tags:${tag}`, key);
      });

      const timeout =
        (options && options.timeout) || this.options.defaultTimeout;
      // Add the data to the key
      if (typeof timeout === "number") {
        multi.set(`data:${key}`, JSON.stringify(data), "EX", timeout);
      } else {
        multi.set(`data:${key}`, JSON.stringify(data));
      }
      const r = await multi.exec();
      this.redis.quit();
      return r;
    } catch (err) {
      return Promise.reject(err);
    }
  };

  // How invalidation by tag works:
  // 1. Get all the keys associated with all the passed-in tags (tags:${tag})
  // 2. Delete all the keys data (data:${key})
  // 3. Delete all the tags (tags:${tag})
  invalidate = async (...tags: Array<string>): Promise<void> => {
    // console.log("tags to clear", tags);
    try {
      // NOTE: [].concat.apply([],...) flattens the array
      // eslint-disable-next-line prefer-spread
      const keys: string[] = [].concat.apply(
        [],
        await Promise.all(
          tags.map((tag) => {
            // console.log(`tags:${tag}`);
            return this.redis.smembers(`tags:${tag}`);
          }),
        ),
      );

      // console.log("keys to clear", keys);
      const pipeline = await this.redis.pipeline();

      keys.forEach((key) => {
        pipeline.del(`data:${key}`);
      });

      tags.forEach((tag) => {
        pipeline.del(`tags:${tag}`);
      });

      await pipeline.exec();
      this.redis.quit();
    } catch (err) {
      return Promise.reject(err);
    }
  };
}

export default TagCache;
